#include <stdio.h>
#include <stdlib.h>

// Algorithm A2.1 (page 68)
int find_span(int n, int p, float u, float* U);

// Algorithm A2.2 (page 70)
void basis_funs(int i, float u, int p, float* U, float* N);